<?php

namespace Autoloading
{

	use \Autoloading\Exception as Exception;

	/**
	 * Automatically loads classes and interfaces when required.
	 * @example:
	 *
	 *    # The only class you have to require "by hand" is the Loader
	 *    require_once 'lib/Autoloading/Loader.class.php';
	 *
	 *    # Next step is to instance a Loader (is a singleton so only one autoloader is allowed). The parameter passed with value TRUE indicates to the loader that it must register itself with the PHP SPL as autoloader
	 *    $Loader = Autoloading\Loader::getInstance(true);
	 *
	 *    # Then you must add one or more directories to search classes and interfaces in. The directories are parsed once, and the result is a map where we store the name of the class or interface and the location in the filesystem. This map is cached to improve efficiency in the next call
	 *    $Autoload->addPath('/path/where/classes/are/');
	 *
	 *    # Then you just can use your classes as if it where required manually
	 *    new \One\Class\Automatically\Loaded();
	 *
	 *    # Or add more directories to map
	 *    $Autoload->addPath('/another/path/');
	 *->addPath('/another/path/');
	 * ->addPath('/another/path/');
	 *    new \Another\Automatically\Loaded\Class\Or\Interface();
	 *
	 *    # You can change the path where maps are stored
	 *    $Autoload->setMapsPath('/var/www/app/cache/autoload/');
	 *
	 * @todo Optimize situation when a parent of a directory already mapped is added
	 * @todo Do not write maps when cache is not used
	 * @todo Expose a method to clean the cache
	 * @todo Let configure the way we serialize the path of directories mapped (can be urlencode, crc32, md5,... )
	 * @todo When getFilesWithClass was implemented, we impacted performance having to use self::arrayMergeRecursiveUnique instead of just array_merge
	 * @todo Store maps in shared memory / memcache / xcache
	 * @todo Change default maps path to the system temporary directory (prioritize convention over configuration)
	 *
	 * @package Autoloading
	 * @author miguel.trias, fipar
	 * @version 2009-Set-01
	 */
	class Loader
	{

		/**
		 * Map caching in filesystem
		 */
		const MAPS_STORAGE_TYPE_FILESYSTEM = 'filesystem';

		/**
		 * Maps storage in Memory
		 */
		const MAPS_STORAGE_TYPE_MEMORY = 'memory';

		/**
		 * Maps storage in nowhere
		 */
		const MAPS_STORAGE_TYPE_NONE = 'no-storage';

		/**
		 * Map files filter : Include
		 */
		const FILTER_INCLUDE = '+';

		/**
		 * Map files filter : Exclude
		 *
		 */
		const FILTER_EXCLUDE = '-';

		/**
		 * Maps Format : PHP parseable Code
		 */
		const MAP_FORMAT_CODE = 'php-code';

		/**
		 * Maps Format : PHP Serialized
		 */
		const MAP_FORMAT_SERIAL = 'php-serial';

		/**
		 * Maps Format : PHP ini
		 */
		const MAP_FORMAT_INI = 'php-ini';

		/**
		 * Default maps path.
		 * This is a relative path (relative to this file) where maps are stored if not specified
		 */
		const DEFAULT_MAPS_DIRECTORY = 'map';

		/**
		 * Autonatically called method
		 */
		const AUTOLOAD_METHOD = 'autoloadClass';

		/**
		 * Maps files extension
		 */
		const MAP_FILES_EXTENSION = 'map';

		/**
		 * The unique instance of this class
		 * @var Loader
		 */
		protected static $instance = null;

		/**
		 * Runtime memory map
		 * @internal this map is constructed in runtime as a merge of all the maps loaded
		 * @var array
		 */
		private $map = array();

		/**
		 * Loaded map files
		 * @internal Unidimensional array with already loaded directories and the correspondant map file
		 * @var array
		 */
		protected $loadedMaps = array();

		/**
		 * Indicates if this instance is already registered
		 * @var boolean
		 */
		private $registered = false;

		/**
		 * Path of maps storage directory (only applicable when in filesystem)
		 * @see setMapsPath()
		 * @var string
		 */
		private $mapsPath = false;

		/**
		 * Filters to apply to files in directories requested to be parsed.
		 *
		 * @var array
		 */
		private $filters = array();

		/**
		 * Cache of map files locations.
		 * Used internally to optimize getPathOfMapFile() method

		 * @see getPathOfMapFile
		 * @var array
		 */
		private $pathOfMapFilesCache = array();

		/**
		 * Maps storage format. Please choose one of self::MAP_FORMAT_* constants if changed
		 *
		 * @var string
		 */
		private $mapCachingFormat = self::MAP_FORMAT_CODE;

		/**
		 * Using maps cache is enabled or not
		 *
		 * @var boolean
		 */
		private $useMapsCache = true;

		/**
		 * Loaded classes
		 * @var array
		 */
		private $loaded_classes = array();

		/**
		 * Loads a file where class or interface $classname is defined
		 * This method is meant to be called automatically by PHP itself
		 * @param string $classname class or interface to be loaded
		 * @return bool indicating success or failure
		 */
		public function autoloadClass($classname)
		{
			if ( !$this->map || !is_array($this->map) || count($this->map) == 0 )
			{
				$this->loadMaps();
			}
			return $this->loadClass($classname);
		}

		/**
		 * Home made implementation of \array_merge_recursive that eliminates duplicated elements
		 * @param array $array1
		 * @param array $array2
		 * @return array
		 */
		private static function arrayMergeRecursiveUnique(array $array1, array $array2)
		{
			$result = $array1;
			foreach($array2 as $key => $value) {
					if(!isset($result[$key]))
					{
							$result[$key] = $value;
					}
					else
					{
							if(!is_array($value)){
									if(is_array($array1[$key]))
									{
											if(!in_array($value,$array1[$key]))
											{
													$result[$key][] = $value;
											}
									}
									else
									{
											if($result[$key] != $value)
													$result[$key] = array($result[$key], $value);
									 }
							}
							else
							{
									$result[$key] = self::arrayMergeRecursiveUnique($result,$array2[$key]);
							}

					}

			}
				unset($key, $value);
				return $result;
		}

		/**
		 * Declares a directory to search classes or interfaces in
		 * @param string $path
		 * @return Loader (returns this instance so you can add more directories in the same line of code)
		 */
		public function addPath($path)
		{
			if ( is_dir($path) )
			{
				self::createMapFile($path);
				$this->loadMapOfPath($path);
				return $this;
			} else {
				self::includeExceptionFile();
				throw new Exception\PathIsNotADirectory($path);
			}
		}

		/**
		 * @return Loader
		 * @param bool $autoregister Indicates you want this instance to be registered in SPL as a class autoloader
		 * @internal This class is a singleton because I don't see any use for multiple instances, althoug the class should work if you make it not single
		 */
		public static function getInstance($autoregister=true)
		{
			if ( !(self::$instance instanceof self) )
			{
				self::$instance = new self($autoregister);
			}
			return self::$instance;
		}

	  /**
     * Loads a class given its FQN (with or without the first slash we assume is a FQN)
     * @param string $classname
     * @return bool
     */
		private function loadClass($classname)
		{
			if ($classname{0} != '\\') {
					$classname = '\\' . $classname;
			}
			if ( ! class_exists ($classname,FALSE) )
			{
				if ( isset($this->map[$classname]) )
				{
					 $classFile = ( is_array($this->map[$classname]) )
											? current($this->map[$classname])
											: $this->map[$classname];
					 array_push($this->loaded_classes, $classname);
					 require_once $classFile;
					 return true;
				} else {
					return false;
				}
				return false;
			} else return true;
		}

		/**
		 * Returns an array with the paths of files where $className is defined.
		 * This normally will return an array with only one element, but there are cases where can be classes declared in multiple files, for example in
		 * @var $className string
		 * @return array
		 */
		public function getFilesWithClass($className)
		{
			$return = array();
			$map = (isset($this->map[$className])) ?: $this->map['\\'.$className];
			if ( $map )
			{
				if ( is_array($map) ) $return = &$map;
				else                  $return = array($map);
			}
			return $return;
		}

		/**
		 * Indica si esta instancia fue registrada como autoload
		 * Already registered in PHP SPL as autoloader
		 * @return boolean
		 */
		public function isRegistered()
		{
			return $this->registered;
		}

		/**
		 * Sets property $this->registered
		 * @return boolean
		 * @see $registered
		 */
		protected function setRegistered()
		{
			$this->registered = true;
		}

		/**
		 * Registers self::AUTOLOAD_METHOD method of this instance as autoloader
		 * @return bool
		 * @throws Exception\Cantregister
		 */
		public function register()
		{
			if ( !$this->isRegistered() )
			{
				if ( !spl_autoload_register ( array($this,self::AUTOLOAD_METHOD) ) )
				{
					self::includeExceptionFile();
					throw new Exception\Cantregister();
				}
				$this->registered = true;
				return true;
			} else {
				self::includeExceptionFile();
				throw new Exception\Cantregister();
			}
		}

		 /**
		  * Unregisters this instance as autoloader if this instance is registered
			* @return bool
			* @throws Exception\Cantunregister
			*/
		public function unregister()
		{
			if ( $this->isRegistered() )
			{
				if ( ! spl_autoload_unregister ( array($this, self::AUTOLOAD_METHOD) ) )
				{
					self::includeExceptionFile();
					throw new Exception\Cantunregister();
				}
				$this->registered = false;
				return true;
			} else {
				self::includeExceptionFile();
				throw new Exception\Cantunregister();
			}
		}

		/**
		 * Sets format for maps caching. Must be one of self::MAP_FORMAT_* constants
		 *
		 * @param string $format
		 */
		public function setMapsCacheFormat($format)
		{
			if ( in_array ( $format,array(self::MAP_FORMAT_INI,self::MAP_FORMAT_CODE,self::MAP_FORMAT_SERIAL) ) )
			{
				$this->mapCachingFormat = $format;
				return true;
			} else {
				self::includeExceptionFile();
				throw new Exception\Invalidmapcachingformat($format);
			}
		}

		/**
		 * Returns maps caching path (when in filesystem)
		 * @return string
		 */
		public function getMapsPath()
		{
			if ( !empty($this->mapsPath) ) return $this->mapsPath;
			else
			{
				self::includeExceptionFile();
				throw new Exception\Mapspathnotset();
			}
		}

		/**
		 * Sets maps caching path (when in filesystem)
		 * @return Loader returns this instance
		 * @param $directory string path to store maps cache
		 * @throws Exception\CantCreateMapsDirectory
		 * @throws Exception\CantWriteMapsDirectory
		 */
		public function setMapsCachingPath($directory)
		{
			 # If directory exists or can be created update the property
			 $this->createDirectory($directory);
			 $this->mapsPath = $directory;

			 # Cleans cache of maps file paths
			 $this->pathOfMapFilesCache = array();

			 return $this;
		}

		/**
		 * Sets conditions to filter paths of files to parse
		 * @return Loader returns this instance
		 * @param array $include
		 * @param array $exclude
		 */
		public function setFilter(array $include, array $exclude = array())
		{
			 $this->filters = array();
			 foreach ($include as $v) $this->filters[self::FILTER_INCLUDE][] = $v;
			 foreach ($exclude as $v) $this->filters[self::FILTER_EXCLUDE][] = $v;

			 return $this;
		}


		/**
		 * Initializes loaded classes array
		 * Sirve para utilizarse desde herramientas (como Deployer.cli.php) que quieran procesar con una sola instancia muchos paths posibles de ejecuci�n
		 */
		public function initLoadedClasses()
		{
				unset($this->loaded_classes);
				$this->loaded_classes = array();
		}

		/**
		 * Generates a list of "require_once" sentences needed if you want/can not to use autoload in some environment
		 */
		public function getDeploymentRequires()
		{
				 $deploymentRequire = '';
				 foreach ($this->loaded_classes as $classname) {
					if (isset($this->map[$classname])) {
							if (!is_array($this->map[$classname])) {
									$deploymentRequire .= 'require_once "' . $this->map[$classname] . '";' . "\n";
							} else {
									$deploymentRequire .= 'require_once "' . $this->map[$classname][0] . '";' . "\n";
							}
					}
				 }
				 return $deploymentRequire;
		}

		/**
		 * Returns a boolean indicating if maps caching is enabled
		 * @see self::$useMapsCache
		 * @return boolean
		 */
		public function getUseMapsCache()
		{
			return $this->useMapsCache;
		}

		/**
		 * Enable or disable using of maps caching
		 *
		 * @param boolean $useOrNot
		 */
		public function setUseMapsCache($useOrNot)
		{
			$this->useMapsCache = (boolean)$useOrNot;
		}

		/**
		 * Includes Autoloading\Exception file.
		 * @internal This file must be included (using this method) before throwing any exception
		 */
		private static function includeExceptionFile()
		{
			if ( !class_exists(__CLASS__.'Exception') )
			{
				require_once __DIR__.DIRECTORY_SEPARATOR . 'Exception.class.php';
			}
		}

		/**
		 * Serializes a path to be used as a map cache filename
		 *
		 * @param string $path
		 * @return string
		 */
		private static function serializePath($path)
		{
			return crc32($path);
		}

		/**
		 * Serializes a map array using the serialization format declared in $this->mapCachingFormat
		 * @see $mapCachingFormat
		 * @param array $map
		 * @return string
		 */
		private function serializeMap(array $map)
		{
			switch ($this->mapCachingFormat)
			{
				 case self::MAP_FORMAT_CODE   :
						return var_export($map,true);
						break;
				 case self::MAP_FORMAT_INI    :
						return self::array2ini($map);
						break;
				 case self::MAP_FORMAT_SERIAL :
						return serialize($map);
						break;
			}
		}

		/**
		 * Unserializes a map array using the inverse serialization format declared in $this->mapCachingFormat
		 *
		 * @param string $map
		 * @return array
		 */
		private function unSerializeMap($map)
		{
			switch ($this->mapCachingFormat)
			{
				 case self::MAP_FORMAT_CODE   :
						$command = "\$tmp = $map;";
						eval($command);
						return $tmp;
						break;

				 case self::MAP_FORMAT_INI    :
						return parse_ini_string($map);
						break;

				 case self::MAP_FORMAT_SERIAL :
						return unserialize($map);
						break;
			}
		}

		/**
		 * Loads the map of the path $path if isnt loaded yet
		 * @param string $path
		 * @return void
		 * @throws Exception\CantReadFile
		 */
		private function loadMapOfPath($path)
		{
			if ( is_string($path) && !empty($path) )
			{
				$mapFile  = $this->getPathOfMapFile($path);
				$pathHash = self::serializePath($path);
				if ( !isset($this->loadedMaps[$path]) )
				{
					 $this->map = self::arrayMergeRecursiveUnique($this->map, $this->getMapOfPath($path));
					 //$this->map = array_merge_recursive($this->map,$this->getMapOfPath($path));
					 $this->loadedMaps[$path] = $mapFile;
					 return true;
				}
			} else {
				self::includeExceptionFile();
				throw new Exception\PathIsNotADirectory($path);
			}
		}

		/**
		 * Reads and unserializes the map cache file of path $path
		 *
		 * @param string $path
		 * @return array
		 */
		private function getMapOfPath($path)
		{
			$mapFile = $this->getPathOfMapFile($path);
			if ( file_exists($mapFile) && is_file($mapFile) && is_readable($mapFile) )
			{
				return ( $this->unserializeMap(file_get_contents( $mapFile )) );
			} else {
				self::includeExceptionFile();
				throw new Exception\CantReadFile($mapFile);
			}
		}

		/**
		 * Loads all maps of specified directories
		 */
		private function loadMaps()
		{
			$this->map = array();
			foreach ( $this->loadedMaps as $scannedDirectory => $mapCacheFile )
			{
				$this->loadMapOfPath($scannedDirectory);
			}
		}

		/**
		 * Returns path to the map cache file of the $path directory
		 *
		 * @param string $directory
		 * @return string
		 */
		private function getPathOfMapFile($directory)
		{
			if ( !isset($this->pathOfMapFilesCache[$directory]) )
			{
				 try
				 {
						$mapsPath = $this->getMapsPath();
				 } catch ( Exception\Mapspathnotset $Ex ) {
						$mapsPath = $this->setMapsCachingPathToDefault();
				 }
				 $this->pathOfMapFilesCache[$directory] = $mapsPath . DIRECTORY_SEPARATOR
																					 . self::serializePath($directory) . '.' . self::MAP_FILES_EXTENSION
																					 ;
			}
			return $this->pathOfMapFilesCache[$directory];
		}

		/**
		 * @param bool $autoregister if true this instance autoregisters itself as autoload
		 * @return void
		 * @see getInstance()
		 */
		protected function __construct($autoregister=true)
		{
			$this->initialize($autoregister);
		}

		/**
		 * Initializes this instance with default configuration
		 * @param boolean $autoregister if true this instance autoregisters itself as autoload
		 */
		private function initialize($register=true)
		{
			$this->setFilter ( array( '.class.php', '.interface.php' )
											 , array( '.svn' )
											 );

			$this->setMapsCacheFormat(self::MAP_FORMAT_CODE);

			if ( $register ) $this->register();
		}

		/**
		 * Sets maps cache path to default value
		 * @return string el mapsPath recien seteado
		 */
		private function setMapsCachingPathToDefault()
		{
			 $mapsPath = dirname(__FILE__)
								 . DIRECTORY_SEPARATOR . self::DEFAULT_MAPS_DIRECTORY
								 ;
			 $this->setMapsCachingPath ( $mapsPath );
			 return $mapsPath;
		}

		/**
		 * Creates maps cache directory if not exists
		 * @return boolean
		 * @throws Exception\CantCreateMapsDirectory
		 */
		private function createMapsCacheDirectory()
		{
			return ( self::createDirectory( $this->getMapsPath() ) );
		}

		/**
		 * Creates a directory with writing permission and returns true on success
		 *
		 * @param string $directory
		 * @return boolean
		 * @throws Exception\CantCreateMapsDirectory
		 */
		private function createDirectory($directory)
		{
			if ( !file_exists($directory) || !is_dir($directory) )
			{
				if ( ! mkdir($directory, 0777, true) )
				{
					self::includeExceptionFile();
					throw new Exception\CantCreateMapsDirectory($directory);
				}
			}
			return true;
		}

		/**
		 * Returns true if the file $file matches the filters and must be parsed
		 *
		 * @param $file nombre de archivo
		 * @return boolean
		 */
		private function fileMatchesFilter($file)
		{

			foreach ( $this->filters[self::FILTER_EXCLUDE] as $filter )
			{
				 if ( false !== stripos($file,$filter) ) return false;
			}

			foreach ( $this->filters[self::FILTER_INCLUDE] as $filter )
			{
				 if ( false !== stripos($file,$filter) ) return true;
			}

			return false;
		}

		/**
		 * Returns a boolean that indicates if one file has declaration of namespaces or not
		 * @param $file string
		 * @return bool
		 */
		private function fileHasNameSpace($file)
		{
			if (! is_readable($file)) {
				self::includeExceptionFile();
				throw new Exception\CantReadFile("No se puede leer el archivo $file");
			}
			$contents = file_get_contents($file);
			preg_match_all("/^namespace\s+(?P<namespace>\w+)\s*[;{]?/mis",$contents,$matches);
			$result = sizeof($matches['namespace'])>0;
			unset($contents);
			unset($matches);
			return $result;
		}

		/**
		 * Creates a map file for the directory $directory
		 * @return bool
		 */
		private function createMapFile($directory)
		{
			$mapFile = $this->getPathOfMapFile($directory);
			if ( !$this->useMapsCache || !file_exists($mapFile) )
			{
				$files = array();
				$dir = new \RecursiveIteratorIterator(new \RecursiveDirectoryIterator($directory),true);

				foreach ($dir as $file => $object)
				{
					if ( is_file($file) && is_readable($file) && $this->fileMatchesFilter($file) )
					{
						if ($this->fileHasNameSpace($file))
						{
							// proceso archivo con declaraci�n de namespaces
							preg_match_all('~^namespace\s+(?P<namespace>[^\s{;]+)\s*[;{]?~mis', file_get_contents($file), $namespacesnames);
							$arrnamespaces = preg_split("/namespace /i",file_get_contents($file));
							$i = 1;
							foreach ($namespacesnames['namespace'] as $namespace)
							{
								$namespace = trim($namespace);
								$nsstr = $arrnamespaces[$i++];
								preg_match_all('~^\s*(?:abstract\s+|final\s+)?(?:class|interface)\s+(?P<class>\w+)~mis', $nsstr, $classes);
								for ($j=0; $j<sizeof($classes['class']); $j++)
								{
									$fqcn = '';
									if (strpos($nsstr, ';') !== FALSE) {
										$fqcn = '\\' . $namespace. '\\' . $classes['class'][$j];
									} else if (strpos($nsstr, '{') !== FALSE) {
										$fqcn = '\\' . $namespace. '\\' . $classes['class'][$j];
									} else {
										$fqcn = '\\' . $namespace. '\\' . $classes['class'][$j];
									}
									# sanity replace
									$fqcn = str_replace("\n", "", $fqcn);
									$newFilePath = realpath($file);
									if ( !isset($files[$fqcn]) )          { $files[$fqcn]   = $newFilePath; }
									elseif ( is_array($files[$fqcn]) )  { $files[$fqcn][] = $newFilePath; }
									elseif ( is_string($files[$fqcn]) ) { $files[$fqcn]   = array($files[$fqcn],$newFilePath);}
									else  { throw new Exception\InvalidDataType('String|Array',$files[$fqcn]); }
								}
							}
						}
						else
						{
							preg_match_all('~^\s*(?:abstract\s+|final\s+)?(?:class|interface)\s+(\w+)~mi', file_get_contents($file), $classes);
							foreach ($classes[1] as $classname)
							{
								$newFilePath = realpath($file);
								$classname = '\\' . $classname;

								/** @todo this should work and be more performant *
								  * $files[$classname] = array_merge_recursive((array)$files[$classname],(array)$newFilePath);
								  */

								if ( !isset($files[$classname]) )          { $files[$classname]   = $newFilePath;
								} elseif ( is_array($files[$classname]) )  { $files[$classname][] = $newFilePath;
								} elseif ( is_string($files[$classname]) ) { $files[$classname]   = array($files[$classname],$newFilePath); // print_r($files);
								} else                                     { throw new Exception\Invaliddatatype('String|Array',$files[$classname]);
								}
							}
						}
					}
				}
				$mapsDirectory = $this->getMapsPath();
				$this->createMapsCacheDirectory();
				if ( touch($mapFile) )
				{
					chmod($mapFile, 0777);
					$map = $this->serializeMap($files);
					file_put_contents($mapFile, $map );
				} else {
					self::includeExceptionFile();
					throw new Exception\CantWriteMapsDirectory($mapsDirectory);
				}
			}

			return true;
		}

		/**
		 * Converts an array to a string suitable for using in an ini file
		 *
		 * @param array $array
		 * @return string
		 */
		private function array2Ini( $array )
		{
			 if (! is_array($array)) {
					return '';
			 }
			 ksort($array);
			 $retorno = '';
			 foreach ($array as $key => $value) {
					if (is_array($value)) {
						 $retorno .= "[$key]\n";
						 $retorno .= array2Ini($value);
					} else {
						 $retorno .= $key . '="' . $value . '"' . "\n";
					}
			 }
			 return $retorno;
		}
	}

}