<?php

/**
 * Unit testing for \Autoloading package
 * Requires PHPUnit.de installed
 * @link htp://phpunit.de/
 * @author miguel.trias
 * @package \Autoloading
 * @version 2009-Set-01
 */

# Bootstrap for PHPUnit
require_once realpath(__DIR__).DIRECTORY_SEPARATOR.'..'.DIRECTORY_SEPARATOR.'init.phpunit.inc';

# Autoload improved for testing
require_once realpath(__DIR__).DIRECTORY_SEPARATOR.'LoaderTester.class.php';

#---------------------------------------------------

class LoaderTest extends PHPUnit_Framework_TestCase
{
    /**
     * One instance to do testing
     * @var LoaderTester
     */
    private $Loader;

    /**
     * Path donde la instancia de autoload testeada almacena los mapas durante el testing
     * @var string
     */
    private $mapsPath;

    public function setUp()
    {
        $this->Loader = \LoaderTester::getInstance(true);

        $this->Loader->setMapsCachingPath(realpath(__DIR__).DIRECTORY_SEPARATOR.'tmp');

        # Stores maps caching directory for use in the tests
        $this->mapsPath = $this->Loader->getMapsPath();

        # Deletes maps cache
        if ( is_dir($this->mapsPath) )
        {
            self::emptyDirectory($this->mapsPath);
            //rmdir($this->mapsPath);
        }
    }

    /**
     * Tests addPath method
     */
    public function testAddPathAdsThisDirectoryToMaps ()
    {
        $tmpPath = __DIR__;
        $this->assertType('\Autoloading\Loader',$this->Loader->addPath($tmpPath),"addPath() method does'nt work for '$tmpPath' directory");
    }

    /**
     * Tests loading classes declared one-per-file
     */
    public function testAddPathLoadsOneClassPerFileClasses()
    {
        $this->assertTrue ( class_exists('ClassOne',true)
                          , 'Error loading classes defined in different files'
						              );
    }

    /**
     * Tests loading of classes defined in the same file
     */
    public function testAddPathLoadsMultipleClassPerFileClasses()
    {
        $this->assertTrue ( class_exists('ClassMultiple1',true) && class_exists('ClassMultiple2',true) && class_exists('ClassMultiple3',true)
            , 'Error loading classes defined in the same file'
        );
    }

    /**
     * Tests that adding a directory twice, does'nt generates another map
     */
    public function testAddPathDoesNotDuplicateSamePath()
    {
        $this->Loader->addPath(__DIR__); $mapsCountPre = self::countFilesIn($this->mapsPath);
        $this->Loader->addPath(__DIR__); $mapsCountPos = self::countFilesIn($this->mapsPath);
        $this->assertEquals($mapsCountPre,$mapsCountPos,"Number of map cache files must be the same when added the same directory twice");
    }

    /**
     * Tests that adding an empty path doesnt generate any map
     * Testing this has any sense ?
     */
    public function testAddPathOfEmptyDirectory()
    {
        $this->Loader->addPath(realpath(__DIR__));
        $mapsCountPre = self::countFilesIn($this->mapsPath);
        $anyDir = dirname(tempnam(DIRECTORY_SEPARATOR.'Directorio_que_no_existe','p'));
    }

    // ---------------------------------------------------------
    // Exceptions testing
    // ---------------------------------------------------------

    /**
     * Trying to add unexistant path produces an exception
     * @expectedException \Autoloading\Exception\PathIsNotADirectory
     */
    public function testAddPathOfUnexistentDirectoryProducesException()
    {
        $this->Loader->addPath(__DIR__.DIRECTORY_SEPARATOR.'popo');
    }

    #---------------------------------------------------------
    // Caching maps testing
    #---------------------------------------------------------

    /**
     * Deleting maps testing
     */
    public function testDeleteMaps()
    {
        $this->Loader->addPath(realpath(__DIR__));
        $this->Loader->deleteLoadedMaps();
        $mapsCount = self::countFilesIn($this->mapsPath);
        $this->assertEquals(0,$mapsCount,"Se ha encontrado algun archivo de mapa cuando se supone que se habian borrado");
    }

    #---------------------------------------------------------
    // Maps cache path changing testing
    #---------------------------------------------------------

    /**
     * Try to change maps cache path
     */
    public function testChangingMapsPath()
    {
			$this->Loader->deleteLoadedMaps();

			$tmpPath = __DIR__.DIRECTORY_SEPARATOR.'tmp';
			$this->Loader->setMapsCachingPath($tmpPath);
			$this->assertEquals( $this->Loader->getMapsPath()
					, $tmpPath
					, "Cant change maps path to '$tmpPath'"
			);
    }

    /**
     * Parse some directory and see if new map cache file is generated in new path
     */
    public function testChangingMapsPathAndAddingPaths()
    {
        $mapsPath = $this->Loader->getMapsPath();
        self::emptyDirectory($mapsPath,true);
        $tmpPath      = dirname(dirname(__FILE__.DIRECTORY_SEPARATOR.'..'));
        $mapsCountPre = self::countFilesIn($mapsPath,true);
        $this->Loader->addPath($tmpPath);
        $mapsCountPos = self::countFilesIn($mapsPath,true);
        $this->assertEquals ( $mapsCountPre
            , $mapsCountPos-1
            , "New map cache file was not generated, $mapsCountPos > $mapsCountPre."
        );
    }

    #---------------------------------------------------------
    // getFilesWithClass method testing
    #---------------------------------------------------------

    /**
     * Pruebo que getFilesWithClass retorne un array
     */
    public function testGetFilesWithClassReturnsAnArray()
    {
        $mapsPath = $this->Loader->getMapsPath();
        self::emptyDirectory($mapsPath);
        $this->Loader->addPath(dirname(realpath(__FILE__)));
        $files = $this->Loader->getFilesWithClass('ClassOne');
        $this->assertType ( 'array'
            , $files
            , "El m�todo getFilesWithClass retorna un array"
        );
    }

    /**
     * Pruebo que getFilesWithClass retorne un array de dos elementos
     */
    public function testGetFilesWithClassReturnsAnArrayWithSomeElements()
    {
        $mapsPath = $this->Loader->getMapsPath();
        self::emptyDirectory($mapsPath);
        $this->Loader->addPath(__DIR__);
        $files = $this->Loader->getFilesWithClass('ClassOne');
        $this->assertEquals ( 2
            , count($files)
            , "El m�todo getFilesWithClass retorna un array con dos elementos para una clase duplicada"
        );
    }

    #---------------------------------------------------------
    // Maps merging test
    #---------------------------------------------------------

    /**
     * Tests to add the same file two times
     */
    public function testNestedIncludedDirectoriesResultsInOneLocationPerClass()
    {
        $mapsPath = $this->Loader->getMapsPath();
        self::emptyDirectory($mapsPath);
        $this->Loader->addPath(__DIR__.DIRECTORY_SEPARATOR.'nested_duplication');
        $this->Loader->addPath(__DIR__.DIRECTORY_SEPARATOR.'nested_duplication'.DIRECTORY_SEPARATOR.'nested_duplication');
        $files = $this->Loader->getFilesWithClass('NestedClass');

        $this->assertType ( 'array'
            , $files
            , "getFilesWithClass always returns an array"
        );
        $this->assertEquals ( 1, count($files)
                            , "getFilesWithClass() doesnt returns only one location for a class added two times. Obtained: ".var_export($files,1)
                            );
        
    }

    #---------------------------------------------------------
    // Testing of different namespace syntax
    #---------------------------------------------------------

		/**
		 * Tests different namespaces declaration sintaxes
		 * IMPORTANT... This classes MUST be defined in separated files for the tests to succeed
		 */
    public function testNamespacesRecomendedSyntax()
    {
        $this->Loader->addPath(__DIR__);
        $this->assertTrue ( class_exists('\NsDefault\Syntax\c8',true) && class_exists('\NsDefault\Syntax\c9',true)
            , 'Loader fails to load classes with namespaces in format (0)'
        );
        $this->assertTrue ( class_exists('\Another\Again\Recomended\Syntax\c5',true) && class_exists('\Another\Again\Recomended\Syntax\c6',true)
            , 'Loader fails to load classes with namespaces in format (1)'
        );
        $this->assertTrue ( class_exists('\Recomended\Syntax\c1',true) && class_exists('\Recomended\Syntax\c2',true)
            , 'Loader fails to load classes with namespaces in format (2)'
        );
        $this->assertTrue ( class_exists('\Another\Recomended\Syntax\c3',true) && class_exists('\Another\Recomended\Syntax\c4',true)
            , 'Loader fails to load classes with namespaces in format (3)'
        );
    }

    #---------------------------------------------------------
    # Tests use of __NAMESPACE__ constant
    #---------------------------------------------------------

    public function testNamespaceConstant()
    {
        $this->Loader->addPath(__DIR__);
        $this->assertTrue ( class_exists('\\Nsp\\Constant\\Test\\n1',true) && class_exists('\\Nsp\\Constant\\Test\\n2',true)
                          , '__NAMESPACE__ constante brakes maps generation'
                          );

				$Instance1 = new \Nsp\Constant\Test\n1();
				$this->assertEquals($Instance1->prop,'Nsp\\Constant\\Test', '__NAMESPACE__ constant brakes maps generation');

				$Instance2 = new \Nsp\Constant\Test\n2();
				$this->assertEquals($Instance2->prop,'Nsp\\Constant\\Test y Algo mas', '__NAMESPACE__ constant brakes maps generation');
    }

    #---------------------------------------------------------
    # Aux Functions
    #---------------------------------------------------------

    /**
     * Counts files in a directory
     * @param string $directory
     * @return integer
     */
    private static function countFilesIn($directory,$ignoreNonExistent=true)
    {
        if ( is_dir($directory) )
        {
            $DI = new DirectoryIterator($directory);
            $count = 0;
            foreach ($DI as $file)
            {
                if ($DI->isDot())
                {continue;}
                else $count++;
            }
            return $count;
        } else if ($ignoreNonExistent) {
            return 0;
        } else {
            throw new Exception("Directorio $directory no existe",__LINE__);
        }
    }

    private static function emptyDirectory($directory,$ignoreNonExistent=true)
    {
        if ( is_dir($directory) )
        {
            $DI = new DirectoryIterator($directory);
            $count = 0;
            foreach ($DI as $file)
            {
                if ($DI->isDot())
                {continue;}
                else
                {
                    $fileToDelete = $directory.DIRECTORY_SEPARATOR.$file;
                    $count++; unlink($fileToDelete);
                }
            }
            return $count;
        } else if ($ignoreNonExistent) {
            return 0;
        } else {
            throw new Exception("Directorio $directory no existe",__LINE__);
        }
    }

#---------------------------------------------------------

}
